import os, requests, pandas as pd
from typing import List

from catrxneng import utils
import catrxneng.quantities as quant
from .step import Step


class Expt:

    @property
    def steady_state_steps(self) -> List[Step]:
        return [step for step in self.steps if step.step_name == "steadyState"]

    @property
    def date_str(self):
        return self.steps[0].start.date_str

    def __init__(
        self,
        expt_type,
        reactor_class,
        kinetic_model_class,
        step_class: type[Step],
        catalyst,
        mcat,
        unit,
        lab_notebook_id,
        project,
        gc=None,
        **kinetic_model_kwargs,
    ):
        self.expt_type = expt_type
        self.reactor_class = reactor_class
        self.kinetic_model_class = kinetic_model_class
        self.kinetic_model_kwargs = kinetic_model_kwargs
        self.step_class = step_class
        self.catalyst = catalyst
        self.mcat = mcat
        self.unit = unit
        self.gc = gc
        self.lab_notebook_id = lab_notebook_id
        self.project = project
        # self.T = T
        # self.p0 = p0
        # self.whsv = whsv
        self.steps: List[Step] = []
        self.pickle_name = self.lab_notebook_id

    def add_step(self, step_name, duration, T, p0, whsv=None, F0=None, start=None):
        if start is None:
            start = self.steps[-1].end
        end = start + duration
        reactor = self.reactor_class(
            T=T,
            kinetic_model_class=self.kinetic_model_class,
            p0=p0,
            whsv=whsv,
            F0=F0,
            mcat=self.mcat,
            **self.kinetic_model_kwargs,
        )
        step = self.step_class(
            self, step_name, len(self.steps) + 1, start, end, reactor
        )
        self.steps.append(step)

    def simulate(self, dt_sec, std_dev=None):
        for step in self.steps:
            step.simulate(dt_sec, std_dev)
        dataframes = [step.time_series_data for step in self.steps]
        self.time_series_data = pd.concat(dataframes, ignore_index=True)
        self._compute_tos()
        for index, step in enumerate(self.steady_state_steps):
            step.steady_state_step_num = index + 1

    def upload_data_to_influx(self, bucket, measurement):
        influx = utils.Influx(
            url=os.getenv("INFLUXDB_URL"),
            org=os.getenv("INFLUXDB_ORG"),
            bucket=bucket,
            measurement=measurement,
        )
        df = self.time_series_data.rename(columns=utils.getconf(self.unit, "tags"))
        influx.upload_dataframe(dataframe=df, token=os.getenv("INFLUXDB_TOKEN"))

    def _compute_tos(self):
        for step in self.steps:
            if step.step_name == "steadyState":
                start = step.start.UET
                break
        self.tos_sec = self.time_series_data["timestamp"] - start
        self.tos_hr = self.tos_sec / 3600.0

    def upload_to_emp(self, host, dt_sec, notes=""):
        endpoint = f"/api/create_simulated_expt/{self.project}"
        url = host + endpoint
        params = {
            "expt_type": self.expt_type,
            "lab_notebook_id": self.lab_notebook_id,
            "unit": self.unit,
            "gc": self.gc,
            "material__common_name": self.catalyst,
            "sample_mass": self.mcat.g,
            "start__ET_str": self.steps[0].start.ET_str,
            "end__ET_str": self.steps[-1].end.ET_str,
            "dt_sec": dt_sec,
            "notes": notes,
        }
        self.delete_from_emp(host)
        resp = requests.post(url, json=params, timeout=10)

        if not resp.ok:
            try:
                error_data = resp.json()
                return {"status_code": resp.status_code, "error": error_data}
            except ValueError:
                return {"status_code": resp.status_code, "body": resp.text}

        try:
            return resp.json()
        except ValueError:
            return {"status_code": resp.status_code, "body": resp.text}

    def delete_from_emp(self, host):
        endpoint = f"/api/delete_expt/{self.project}"
        if not host.startswith("http://") and not host.startswith("https://"):
            host = "http://" + host
        url = host + endpoint
        params = {"lab_notebook_id": self.lab_notebook_id, "project": self.project}
        resp = requests.delete(url, json=params, timeout=10)

        if not resp.ok:
            try:
                error_data = resp.json()
                return {"status_code": resp.status_code, "error": error_data}
            except ValueError:
                return {"status_code": resp.status_code, "body": resp.text}

        try:
            return resp.json()
        except ValueError:
            return {"status_code": resp.status_code, "body": resp.text}

    def compute_inlet_partial_pressures(self):
        df = self.time_series_data.copy()
        mask = [col for col in df.columns if "_mfc" in col]
        total_inlet_flow_molh = df[mask].sum(axis=1)
        for col in mask:
            y = df[col] / total_inlet_flow_molh
            p_col_id = f"p_{col.split('_mfc')[0]}_bar"
            df[p_col_id] = y * df["pressure"]
        return df



import catrxneng.utils.equations as eqn
import catrxneng.quantities as quant
from .species import Species


class CO(Species):
    CLASS = "carbon_oxide"
    C_ATOMS = 1
    O_ATOMS = 1
    MOL_WEIGHT = 28
    HF_298_GAS = quant.Energy(kJmol=-110.53)
    S_298_GAS = quant.Entropy(JmolK=197.66)
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 298,
            "max_temp_K": 1300,
            "phase": "gas",
            "A": 25.56759,
            "B": 6.096130,
            "C": 4.054656,
            "D": -2.671301,
            "E": 0.131021,
            "F": -118.0089,
            "G": 227.3665,
            "H": -110.5271,
        }
    ]

    @classmethod
    def Hf_gas(cls, T):
        return eqn.Hf_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def S_gas(cls, T):
        return eqn.S_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def Cp_gas(cls, T):
        return eqn.Cp_shomate(T, cls.NIST_THERMO_PARAMS[0])



import numpy as np
from numpy.typing import NDArray
from scipy.optimize import least_squares

import catrxneng as cat
import catrxneng.quantities as quant
from catrxneng.species import Species


class KineticModel:
    LIMITING_REACTANT: str
    T_REF: quant.Temperature
    REACTANTS: dict[str, type[Species] | Species]
    PRODUCTS: dict[str, type[Species] | Species]
    COMPONENTS: dict[str, type[Species] | Species]
    REACTIONS: dict[str, type[cat.reactions.Reaction]]
    STOICH_COEFF: NDArray
    ORDER: NDArray
    C_ATOMS: NDArray

    kref: NDArray
    Ea: NDArray

    def __init__(self, T):
        self.T = T
        self.comp_idx = {comp: i for i, comp in enumerate(self.COMPONENTS)}
        self.rxn_idx = {rxn: i for i, rxn in enumerate(self.REACTIONS)}
        self.fugacity_coeff = np.ones(len(self.COMPONENTS))

    def compute_temp_dependent_constants(self):
        raise NotImplementedError(
            "Child class must implement compute_temp_dependent_constants."
        )

    @classmethod
    def comp_list(cls):
        return list(cls.COMPONENTS.keys())

    @property
    def T(self):
        return self._T

    @T.setter
    def T(self, value):
        if value is not None and (not hasattr(self, "_T") or value.si != self._T.si):
            self._T = value
            self.compute_temp_dependent_constants()

    def _compute_final_molfrac(self, initial_moles, extent):
        delta_moles = self.STOICH_COEFF.T @ extent
        final_moles = initial_moles + delta_moles
        total_final_moles = np.sum(final_moles)
        molfrac = final_moles / total_final_moles
        return molfrac, delta_moles

    def _compute_extent_bounds(self, initial_moles):
        lower_bounds = [0] * len(self.REACTIONS)
        upper_bounds = []
        for i in range(len(self.REACTIONS)):
            stoich_lim = self.STOICH_COEFF[i][self.comp_idx[self.LIMITING_REACTANT]]
            max_extent = initial_moles[self.comp_idx[self.LIMITING_REACTANT]] / abs(
                stoich_lim
            )
            upper_bounds.append(max_extent)
        return (lower_bounds, upper_bounds)

    def _equilibrium_objective(self, extent, P_bar, initial_moles, Keq):
        molfrac = self._compute_final_molfrac(initial_moles, extent)[0]
        molfrac[molfrac < 0] = 0.00001
        p = molfrac * P_bar
        fugacity = self.fugacity_coeff * p
        activity = fugacity / quant.STD_STATE_FUGACITY.bar
        K_calc = np.prod(activity**self.STOICH_COEFF, axis=1)
        return (np.log(K_calc) - np.log(Keq)) ** 2

    def equilibrate(
        self,
        p0: quant.Pressure,
        T: quant.Temperature,
        initial_guesses=None,
        allow_component_mismatch=False,
    ):
        self.T = T
        initial_total_moles = 100
        if allow_component_mismatch:
            p0_bar = [p0[comp].bar for comp in self.comp_list()]
            p0 = quant.Pressure(bar=p0_bar, keys=self.comp_list())
        P = quant.Pressure(si=np.sum(p0.si))
        initial_molfrac = p0 / P
        initial_moles = initial_molfrac.si * initial_total_moles
        assert isinstance(initial_moles, np.ndarray)
        if not initial_guesses:
            num_rxns = len(self.REACTIONS)
            initial_guess = (
                initial_moles[self.comp_idx[self.LIMITING_REACTANT]] / num_rxns
            )
            initial_guesses = np.ones(num_rxns) * initial_guess / 2

        def objective(extent):
            return self._equilibrium_objective(extent, P.bar, initial_moles, self.Keq)

        solution = least_squares(
            objective,
            initial_guesses,
            bounds=self._compute_extent_bounds(initial_moles),
            method="trf",
            ftol=1e-10,
            max_nfev=1000,
        )
        extent = solution.x
        eq_molfrac, delta_moles = self._compute_final_molfrac(initial_moles, extent)
        delta = delta_moles[self.comp_idx[self.LIMITING_REACTANT]]
        initial = initial_moles[self.comp_idx[self.LIMITING_REACTANT]]
        self.eq_conversion = quant.Fraction(si=-delta / initial)
        self.eq_molfrac = quant.Fraction(si=eq_molfrac, keys=self.comp_list())
        self.eq_partial_pressure = self.eq_molfrac * P

    @classmethod
    def carbon_balance(cls, reactor):
        carbon_in = np.sum(reactor.F0.si * cls.C_ATOMS)
        carbon_out = np.sum(reactor.F.si * cls.C_ATOMS[:, np.newaxis], axis=0)
        c_bal = cat.utils.divide(carbon_out, carbon_in)
        return quant.Fraction(si=c_bal)

    def calculate_rates(self, p: quant.Pressure):
        raise NotImplementedError("Child class must implement calculate_rates().")



import os
from importlib import import_module
from typing import cast
import numpy as np
from numpy.typing import NDArray
import pandas as pd
from math import log10, floor

from .influx import Influx
from .time import Time


def divide(x, y):
    if isinstance(y, (np.ndarray, pd.Series)):
        y_safe = y.copy()
        y_safe[y_safe == 0] = np.nan
    else:
        y_safe = np.nan if y == 0 else y

    return x / y_safe


def getconf(conf_name, variable):
    conf_module_path = os.getenv("CONF_MODULE_PATH")
    module_path = f"{conf_module_path}.{conf_name}"
    conf_module = import_module(module_path)
    return getattr(conf_module, variable)


def filter_df(df: pd.DataFrame, col: str, range: list) -> pd.DataFrame:
    filtered_df = df[(df[col] > range[0]) & (df[col] < range[1])]
    return cast(pd.DataFrame, filtered_df.reset_index(drop=True))


def get_unique_values(series: pd.Series, tol: float) -> np.typing.NDArray:
    rounded = (series / tol).round() * tol
    return rounded.unique()


def apply_sig_figs(x: float, n: int = 3) -> float:
    if x == 0 or np.isnan(x):
        return x
    rounded = round(x, -int(floor(log10(abs(x)))) + (n - 1))
    if rounded == int(rounded):
        return int(rounded)
    return rounded


def apply_sig_figs_to_df(df: pd.DataFrame, n: int = 3) -> pd.DataFrame:
    float_cols = df.select_dtypes(include=["float64", "float32"]).columns
    df.loc[:, float_cols] = df[float_cols].map(lambda x: apply_sig_figs(x, n))
    return df



from .species import Species
from ..quantities import *

class Graphite(Species):
    CLASS = "solid_carbon"
    C_ATOMS = 1
    MOL_WEIGHT = 12
    S_298 = Entropy(JmolK=5.6)
    
    def Cp(self, T):
        return HeatCapacity(JmolK=8.23)



from catrxneng import utils
from .species import Species
from ..quantities import *


class NH3(Species):
    CLASS = "inorganic_hydride"
    H_ATOMS = 3
    N_ATOMS = 1
    MOL_WEIGHT = 17
    HF_298_GAS = Energy(kJmol=-45.9)
    S_298_GAS = Entropy(JmolK=192.77)
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 298,
            "max_temp_K": 1400,
            "phase": "gas",
            "A": 19.99563,
            "B": 49.77119,
            "C": -15.37599,
            "D": 1.921168,
            "E": 0.189174,
            "F": -53.30667,
            "G": 203.8591,
            "H": -45.89806,
        }
    ]

    # @classmethod
    # def Hf_gas(cls, T):
    #     thermo_params = cls._get_thermo_params(T)
    #     return eqn.Hf_shomate(T, thermo_params)
    @classmethod
    def Hf_gas(cls, T):
        return eqn.Hf_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def S_gas(cls, T):
        thermo_params = cls._get_thermo_params(T)
        return eqn.S_shomate(T, thermo_params)

    @classmethod
    def Cp_gas(cls, T):
        thermo_params = cls._get_thermo_params(T)
        return eqn.Cp_shomate(T, thermo_params)



import numpy as np
from typing import Any
from catrxneng import utils
import catrxneng.quantities as quant


class GasMixture:

    def __init__(
        self,
        components: dict[str, Any],
        y: quant.Fraction | None = None,
        p: quant.Pressure | None = None,
    ):
        self.components = components
        if y and not p:
            self.y = y
        elif p and not y:
            self.p = p
        else:
            raise ValueError("p or y can be specified, not both.")

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        from ..quantities import Fraction

        if not hasattr(self, "_y") or value.si != self._y.si:
            self._y = value
            active_comp_list = self.y.keys.copy()
            try:
                active_comp_list.remove("inert")
            except ValueError:
                pass
            # self.components = {
            #     comp_id: species.CLASS_MAP[comp_id] for comp_id in self.y.keys
            # }
            y_active = Fraction(
                si=[self.y[comp_id].si for comp_id in active_comp_list],
                keys=active_comp_list,
            )
            y_active_sum = np.sum(y_active.si)
            y_active_normalized = y_active / y_active_sum

            active_components = {
                comp_id: component
                for comp_id, component in self.components.items()
                if comp_id.lower() != "inert"
            }
            self.avg_mol_weight = np.sum(
                [
                    component.MOL_WEIGHT * y_active_normalized[comp_id].si
                    for comp_id, component in active_components.items()
                ]
            )

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, value):
        from ..quantities import Pressure

        if not hasattr(self, "_p") or value.si != self._p.si:
            self._p = value
            self.P = Pressure(si=np.sum([self._p.si]))
            self.y = utils.divide(self._p, self.P)



import pandas as pd, numpy as np

from .step import Step
import catrxneng.quantities as quant


class Co2ToMeohStep(Step):

    def simulate(self, dt_sec, std_dev=None):
        super().simulate(dt_sec, std_dev)
        df = self.time_series_data
        df["n2_mfc"] = self.reactor.F0["inert"].smLmin * np.random.normal(
            loc=1, scale=std_dev["mfc"], size=self.num_points
        )
        df["co2_mfc"] = self.reactor.F0["co2"].smLmin * np.random.normal(
            loc=1, scale=std_dev["mfc"], size=self.num_points
        )
        df["h2_mfc"] = self.reactor.F0["h2"].smLmin * np.random.normal(
            loc=1, scale=std_dev["mfc"], size=self.num_points
        )

        self.populate_inlet_partial_pressures()
        df = df.rename(columns={"inert_gc_conc": "n2_gc_conc", "p_inert": "p_n2"})
        df["dme_gc_conc"] = 0.0
        if "ch4_gc_conc" not in df.columns:
            df["ch4_gc_conc"] = 0.0
        self.time_series_data = df



from typing import TYPE_CHECKING

from .quantity import Quantity
from catrxneng.utils import *

if TYPE_CHECKING:
    from ..species import GasMixture


class SpaceTime(Quantity):

    def __init__(self, gas_mixture: "GasMixture" = None, **kwargs):
        self.gas_mixture = gas_mixture
        super().__init__(**kwargs)

    @property
    def skgcatmol(self):
        return self.si

    @skgcatmol.setter
    def skgcatmol(self, value):
        self.si = self.to_float(value)

    @property
    def hgcatmol(self):
        return self.si / 3600 * 1000

    @hgcatmol.setter
    def hgcatmol(self, value):
        self.si = self.to_float(value) * 3600 / 1000

    @property
    def hgcatsmL(self):
        return self.si / 3600 / 22.4

    @hgcatsmL.setter
    def hgcatsmL(self, value):
        self.si = self.to_float(value) * 3600 * 22.4

    @property
    def hr(self):
        try:
            return self.si * 1000 / self.gas_mixture.avg_mol_weight / 3600
        except TypeError:
            raise AttributeError("Spacetime has no gas mixture assigned.")

    @hr.setter
    def hr(self, value):
        try:
            self.si = value / 1000 * self.gas_mixture.avg_mol_weight * 3600
        except TypeError:
            raise AttributeError("Spacetime has no gas mixture assigned.")

    @property
    def sec(self):
        try:
            return self.si * 1000 / self.gas_mixture.avg_mol_weight
        except TypeError:
            raise AttributeError("Spacetime has no gas mixture assigned.")

    @sec.setter
    def sec(self, value):
        try:
            self.si = value / 1000 * self.gas_mixture.avg_mol_weight
        except TypeError:
            raise AttributeError("Spacetime has no gas mixture assigned.")



import os, pandas as pd
import sys
from influxdb_client import InfluxDBClient, WritePrecision  # type: ignore

from catrxneng.utils.time import Time


class Influx:

    def __init__(self, url: str, org: str, bucket: str, measurement: str):
        self.url: str = url
        self.org: str = org
        self.bucket: str = bucket
        self.measurement: str = measurement
        self.raw_dataframes: list[pd.DataFrame]

    def upload_dataframe(self, dataframe: pd.DataFrame, token: str):
        dataframe = dataframe.copy()
        dataframe.set_index("timestamp", inplace=True)

        with InfluxDBClient(url=self.url, token=token, org=self.org) as client:

            with client.write_api() as write_api:
                try:
                    write_api.write(
                        bucket=self.bucket,
                        record=dataframe,
                        data_frame_measurement_name=self.measurement,
                        write_precision=WritePrecision.S,
                    )
                except Exception as e:
                    print(f"Write operation failed with exception: {e}")
                    raise

    def upload_points(self, points, token, error_blurb=None):
        with InfluxDBClient(url=self.url, token=token, org=self.org) as client:
            with client.write_api() as write_api:
                try:
                    write_api.write(bucket=self.bucket, org=self.org, record=points)
                    return True
                except Exception as e:
                    print(f"Failed to write '{error_blurb}' to Influx: {e}")
                    sys.stdout.flush()
                    sys.stderr.flush()
                    return False

    def generate_query(
        self, tags: dict[str, str], start: Time, end: Time, dt_sec: int = 1
    ):
        self.tags = tags
        tag_string = ""
        for value in tags.values():
            tag_string = tag_string + 'r["_field"] == "' + value["tag"] + '" or '
        tag_string = tag_string[:-4]
        tag_string = tag_string + ")"

        query_start = (
            start.UTC.strftime("%Y-%m-%d") + "T" + start.UTC.strftime("%H:%M:%S") + "Z"
        )
        query_end = (
            end.UTC.strftime("%Y-%m-%d") + "T" + end.UTC.strftime("%H:%M:%S") + "Z"
        )

        self.query = """
        from(bucket: "BUCKET")
        |> range(start: START, stop: END)
        |> filter(fn: (r) => r["_measurement"] == "MEASUREMENT")
        |> filter(fn: (r) => TAGS
        |> aggregateWindow(every: SECONDSs, fn: last, createEmpty: false)
        |> map(fn: (r) => ({ r with _time: int(v: r._time) / 1000000000}))
        |> keep(columns: ["_time", "_field", "_value"])
        |> yield(name: "last")
        |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
        """

        self.query = self.query.replace("BUCKET", self.bucket)
        self.query = self.query.replace("MEASUREMENT", self.measurement)
        self.query = self.query.replace("START", query_start)
        self.query = self.query.replace("END", query_end)
        self.query = self.query.replace("TAGS", tag_string)
        self.query = self.query.replace("SECONDS", str(dt_sec))

    def request_data(self, token: str):
        timeout_min = 0.5
        max_retries = 3  # Maximum number of retries
        retry_count = 0

        while retry_count < max_retries:
            with InfluxDBClient(
                url=self.url,
                token=token,
                org=self.org,
                timeout=timeout_min * 60 * 1000,
            ) as client:
                try:
                    self.raw_dataframes = client.query_api().query_data_frame(
                        query=self.query
                    )
                    return  # Exit the method if the query is successful
                except ValueError:
                    # Handle the case where the query fails and needs adjustment
                    self.query = self.query.split("|> pivot")[0]
                    # retry_count += 1
                except Exception as e:
                    # Handle timeout or other exceptions
                    print(
                        f"Query failed with error: {e}. Retrying... ({retry_count + 1}/{max_retries})"
                    )
                    retry_count += 1

        raise TimeoutError("Failed to retrieve data after multiple retries.")

    def format_data(self):
        self.data = pd.concat(self.raw_dataframes, ignore_index=True)
        self.data = self.data[["_time", "_field", "_value"]]
        self.data.rename(
            columns={"_time": "timestamp", "_value": "value"}, inplace=True
        )
        self.data_dict = {
            key: self.data[self.data["_field"] == value][["timestamp", "value"]]
            for key, value in self.tags.items()
        }
        for key, dataset in self.data_dict.items():
            try:
                config = self.unit.tags[key]
                dataset.loc[:, "value"] = dataset["value"] * config.get("multiplier", 1)
                dataset.loc[:, "value"] = dataset["value"] + config.get("add", 0)
            except TypeError:
                pass



import numpy as np
from numpy.typing import NDArray

from ..kinetic_model import KineticModel
import catrxneng.utils.equations as eqn
from catrxneng import reactions
import catrxneng.species as species
import catrxneng.quantities as quant


class Co2ToMeohWithDehydration(KineticModel):
    LIMITING_REACTANT = "co2"
    # T_REF = quant.Temperature(C=400)
    REACTANTS = {
        "co2": species.CO2,
        "h2": species.H2,
    }
    PRODUCTS = {
        "ch3oh": species.CH3OH,
        "h2o": species.H2O,
        "dme": species.DME,
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    REACTIONS = {
        "co2_to_meoh": reactions.Co2ToMeoh,
        "meoh_dehydration": reactions.MeohDehydration,
    }
    # ORDER = np.array([0.78, 1.5])
    STOICH_COEFF = np.array(
        [
            [-1, -3, 1, 1, 0, 0],
            [0, 0, -2, 1, 1, 0],
        ]
    )
    # KREF_MOLSKGCATBAR = np.array([0.000373, 1.8e-3])
    # EA_KJMOL = np.array([100.1, 54.5])

    def __init__(
        self,
        T: quant.Temperature | None = None,
        kref: NDArray | None = None,
        Ea: quant.Energy | None = None,
    ):
        self.kref = kref
        # if self.kref is None:
        #     self.kref = self.KREF_MOLSKGCATBAR
        self.Ea = Ea
        # if self.Ea is None:
        #     self.Ea = self.EA_KJMOL
        super().__init__(T)

    def compute_temp_dependent_constants(self):
        self.Keq_co2_to_meoh = reactions.Co2ToMeoh.Keq(self.T)
        self.Keq_meoh_dehydration = reactions.MeohDehydration.Keq(self.T)
        self.Keq = np.array([self.Keq_co2_to_meoh, self.Keq_meoh_dehydration])



from .species import Species
from ..quantities import Energy, Entropy, Temperature

import catrxneng.utils.equations as eqn


class H2O(Species):
    CLASS = "polar_solvent"
    H_ATOMS = 2
    O_ATOMS = 1
    MOL_WEIGHT = 18
    HF_298_LIQ = Energy(kJmol=-285.83)
    S_298_LIQ = Entropy(JmolK=69.95)
    HF_298_GAS = Energy(kJmol=-241.83)
    S_298_GAS = Entropy(JmolK=188.84)
    BOILING_TEMP = Temperature(C=100)
    # DH_VAP = Energy(kJmol=40.657)
    # DS_VAP = Entropy(JmolK=109)
    DH_VAP = HF_298_GAS - HF_298_LIQ
    DS_VAP = S_298_GAS - S_298_LIQ
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 298,
            "max_temp_K": 500,
            "phase": "liq",
            "A": -203.606,
            "B": 1523.29,
            "C": -3196.413,
            "D": 2474.455,
            "E": 3.855326,
            "F": -256.5487,
            "G": -488.7163,
            "H": -285.8304,
        },
        {
            "min_temp_K": 500,
            "max_temp_K": 1700,
            "phase": "gas",
            "A": 30.09200,
            "B": 6.832514,
            "C": 6.793435,
            "D": -2.534480,
            "E": 0.082139,
            "F": -250.8810,
            "G": 223.3967,
            "H": -241.8264,
        },
    ]

    @classmethod
    def Hf_gas(cls, T):
        return eqn.Hf_shomate(T, cls.NIST_THERMO_PARAMS[1])

    @classmethod
    def S_gas(cls, T):
        return eqn.S_shomate(T, cls.NIST_THERMO_PARAMS[1])

    @classmethod
    def Cp_gas(cls, T):
        from ..quantities import HeatCapacity

        params = cls.NIST_THERMO_PARAMS[1]
        if params["min_temp_K"] <= T.K <= params["max_temp_K"]:
            return eqn.Cp_shomate(T, params)
        elif 298 <= T.K <= 500:
            return HeatCapacity(JmolK=34.52)
        else:
            raise ValueError("Invalid temperature for H2O Cp_gas.")

    @classmethod
    def Cp_liq(cls, T):
        params = cls.NIST_THERMO_PARAMS[0]
        if params["min_temp_K"] <= T.K <= params["max_temp_K"]:
            return eqn.Cp_shomate(T, params)
        else:
            raise ValueError("Invalid temperature for H2O Cp_liq.")



from .ghosh2021 import Ghosh2021
from .ghosh2021_without_sabatier import Ghosh2021WithoutSabatier
from .ghosh2021_no_h2_in_denom_rev import Ghosh2021WithoutSabatierRev
from .power_law import PowerLaw
from .co2_to_meoh_with_dehydration import Co2ToMeohWithDehydration



import numpy as np
from .co2_to_meoh import Co2ToMeoh
from .reaction import Reaction
from ..quantities import *
from ..species import *


class MeohToCo2(Reaction):
    COMPONENTS = {
        "ch3oh": CH3OH,
        "h2o": H2O,
        "co2": CO2,
        "h2": H2,
        "inert": Ar,
    }
    STOICH_COEFF = Dimensionless(
        si=[-1.0, -1.0, 1.0, 3.0, 0.0], keys=list(COMPONENTS.keys())
    )
    DEFAULT_LIMITING_REACTANT = "ch3oh"

    @staticmethod
    def Keq(T):
        return np.reciprocal(Co2ToMeoh.Keq(T))




from typing import Any
from numpy.typing import NDArray
from .quantity import Quantity
from catrxneng.utils import *


class Pressure(Quantity):

    # def __init__(self, **kwargs):
    #     if len(kwargs.keys()) == 1 and kwargs.get("keys", None) is not None:
    #         si = {"si": np.zeros(len(kwargs["keys"]))}
    #         kwargs = {**si, **kwargs}
    #     super().__init__(**kwargs)

    @property
    def Pa(self):
        return self.si

    @Pa.setter
    def Pa(self, value):
        self.si = self.to_float(value)

    @property
    def bar(
        self,
    ) -> pd.Series | NDArray[np.number[Any]] | float:
        return self.si / 100000

    @bar.setter
    def bar(self, value):
        self.si = self.to_float(value) * 100000

    @property
    def kPa(self):
        return self.si / 1000

    @kPa.setter
    def kPa(self, value):
        self.si = self.to_float(value) * 1000

    @property
    def atm(self):
        return self.si / 101325

    @atm.setter
    def atm(self, value):
        self.si = self.to_float(value) * 101325

    @property
    def MPa(self):
        return self.si / 1000000

    @MPa.setter
    def MPa(self, value):
        self.si = self.to_float(value) * 1000000



import numpy as np
from scipy.integrate import solve_ivp

from .. import utils
from .reactor import Reactor
from catrxneng.kinetic_models import KineticModel
from catrxneng.species.gas_mixture import GasMixture
import catrxneng.quantities as quant


class PFR(Reactor):
    def __init__(
        self,
        kinetic_model_class: type[KineticModel],
        T: quant.Temperature,
        p0: quant.Pressure,
        whsv: quant.WHSV | None = None,
        mcat: quant.Mass | None = None,
        F0: quant.MolarFlowRate | None = None,
        **kinetic_model_kwargs,
    ):
        super().__init__()
        self.kinetic_model_class = kinetic_model_class
        self.kinetic_model: KineticModel = kinetic_model_class(
            T=T, **kinetic_model_kwargs
        )
        self.T = T
        self.p0 = p0
        feed = GasMixture(components=self.kinetic_model_class.COMPONENTS, p=self.p0)
        self.P = quant.Pressure(si=np.sum(p0.si))
        self.P_bar = self.P.bar
        self.y0 = utils.divide(p0, self.P)
        self.whsv = whsv
        self.mcat: quant.Mass = quant.Mass(g=1) if mcat is None else mcat
        if self.whsv:
            if self.whsv.gas_mixture is None:
                self.whsv.gas_mixture = feed
            self.Ft0_active = self.whsv * self.mcat
            self.Ft0 = self.Ft0_active / (1 - self.y0["inert"])
            self.F0 = self.y0 * self.Ft0
        elif F0 is not None:
            self.F0 = F0
            self.Ft0 = quant.MolarFlowRate(molh=np.sum(self.F0.molh))
            self.Ft0_active = self.Ft0 - self.F0["inert"]
            self.whsv = quant.WHSV(
                molhgcat=self.Ft0_active.molh / self.mcat.g, gas_mixture=feed
            )
        else:
            raise ValueError("whsv and F0 cannot both be None.")
        self.check_components()

    def dFdw(self, x, F: np.ndarray) -> np.array:
        p_array = F / np.sum(F) * self.P_bar
        return self.kinetic_model.rate_equations(p_array)

    def dfdx(self, x, f: np.ndarray) -> np.array:
        Ft = np.sum(self.Ft0_active.si * f)
        p = self.P.si * self.Ft0_active.si * f / Ft
        p = quant.Pressure(si=p, keys=self.p0.keys.copy())
        return (
            self.mcat.si
            / self.Ft0_active.si
            * np.array(
                [
                    rate(p, self.T).si
                    for rate in self.kinetic_model.rate_equations.values()
                ]
            )
        )

    def _solve_dimensional(self, points: int, method: str):
        w_span = (0, self.mcat.g)
        w_eval = np.linspace(0, self.mcat.g, points)
        F0 = self.F0.molh
        solution = solve_ivp(self.dFdw, w_span, F0, t_eval=w_eval, method=method)
        self.w = quant.Mass(g=solution.t)
        self.F = quant.MolarFlowRate(molh=solution.y, keys=self.F0.keys)

    def _solve_dimensionless(
        self, points: int, method: str, rtol: float = 1e-3, atol: float = 1e-6
    ):
        x_span = (0, 1)
        x_eval = np.linspace(x_span[0], x_span[1], points)
        f0 = self.F0.si / self.Ft0_active.si
        solution = solve_ivp(
            self.dfdx, x_span, f0, t_eval=x_eval, method=method, rtol=rtol, atol=atol
        )
        self.x = solution.t
        self.f = solution.y
        self.w = quant.Mass(si=(self.x * self.mcat.si))
        F = self.Ft0_active.si * self.f
        self.F = quant.MolarFlowRate(si=F, keys=self.F0.keys)

    def solve(
        self, points: int = 1000, nondimensionalize: bool = False, method: str = "LSODA"
    ):
        if nondimensionalize:
            self._solve_dimensionless(points=points, method=method)
        else:
            self._solve_dimensional(points=points, method=method)
        y_si = np.divide(self.F.si, np.sum(self.F.si, axis=0))
        self.y = quant.Fraction(si=y_si, keys=self.F0.keys)
        self.conversion = utils.divide(
            self.F0[self.kinetic_model.LIMITING_REACTANT]
            - self.F[self.kinetic_model.LIMITING_REACTANT],
            self.F0[self.kinetic_model.LIMITING_REACTANT],
        )
        spacetime = utils.divide(self.w.g, self.Ft0_active.smLh)
        self.spacetime = quant.SpaceTime(hgcatsmL=spacetime)
        self.whsv_array = quant.WHSV(smLhgcat=utils.divide(1, spacetime))
        vol_flow_rate = self.Ft0.si * quant.R.si * self.T.si / self.P.si
        self.vol_flow_rate = quant.VolumetricFlowRate(si=vol_flow_rate)
        self.p = self.y * self.P
        self.rate = self.kinetic_model.calculate_rates(self.p)

    @property
    def conv_rel_to_eq_conv(self):
        self.kinetic_model.equilibrate(p0=self.p0, T=self.T)
        return self.conversion[-1] / self.kinetic_model.eq_conversion

    def compute_carbon_basis_selectivity(self):
        prod_carbon_molh = np.zeros(self.F.si.shape[1])
        for species_id, species in self.kinetic_model_class.PRODUCTS.items():
            prod_carbon_molh += self.F[species_id].molh * species.C_ATOMS
        sel = np.array(
            [
                self.F[species_id].molh * species.C_ATOMS / prod_carbon_molh  # type: ignore
                for species_id, species in self.kinetic_model_class.PRODUCTS.items()
            ]
        )
        self.carbon_basis_selectivity = quant.Fraction(
            si=sel, keys=list(self.kinetic_model_class.PRODUCTS.keys())
        )



"""
Simple utilities for project setup.
"""
import sys
from pathlib import Path


def setup_project_path():
    """
    Add the project root to sys.path so catrxneng can be imported from anywhere.
    """
    # Find the catrxneng directory by walking up
    current = Path.cwd()
    while current.name != 'catrxneng' and current.parent != current:
        current = current.parent

    if current.name == 'catrxneng':
        project_root = current.parent
        if str(project_root) not in sys.path:
            sys.path.insert(0, str(project_root))
        return project_root

    # Fallback: assume we're somewhere under the project
    # Try going up 2-3 levels
    for levels in range(1, 4):
        candidate = Path.cwd()
        for _ in range(levels):
            candidate = candidate.parent
        if (candidate / 'catrxneng').exists():
            if str(candidate) not in sys.path:
                sys.path.insert(0, str(candidate))
            return candidate

    return None


import catrxneng.utils.equations as eqn
import catrxneng.quantities as quant
from .species import Species


class CO2(Species):
    CLASS = "carbon_oxide"
    C_ATOMS = 1
    O_ATOMS = 2
    MOL_WEIGHT = 44
    HF_298_GAS = quant.Energy(kJmol=-393.51)
    S_298_GAS = quant.Entropy(JmolK=213.79)
    NIST_THERMO_PARAMS = [
        {
            "min_temp_K": 298,
            "max_temp_K": 1200,
            "phase": "gas",
            "A": 24.99735,
            "B": 55.18696,
            "C": -33.69137,
            "D": 7.948387,
            "E": -0.136638,
            "F": -403.6075,
            "G": 228.2431,
            "H": -393.5224,
        }
    ]

    @classmethod
    def Hf_gas(cls, T):
        return eqn.Hf_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def S_gas(cls, T):
        return eqn.S_shomate(T, cls.NIST_THERMO_PARAMS[0])

    @classmethod
    def Cp_gas(cls, T):
        return eqn.Cp_shomate(T, cls.NIST_THERMO_PARAMS[0])



from .dimensionless import Dimensionless


class Fraction(Dimensionless):

    @property
    def pct(self):
        return self.si * 100

    @pct.setter
    def pct(self, value):
        self.si = value / 100



import plotly.graph_objects as go
import pandas as pd

from catrxneng.kinetic_models import KineticModel


class Reactor:

    kinetic_model_class: type[KineticModel]

    @property
    def molar_flowrate_df_molh(self):
        try:
            return self.molar_flowrate_df_molh_cache
        except AttributeError:
            molar_flowrate_dict = {
                key: value for key, value in zip(self.F.keys, self.F.molh)
            }
            self.molar_flowrate_df_molh_cache = molar_flowrate_dict
            return self.molar_flowrate_df_molh_cache

    @molar_flowrate_df_molh.setter
    def molar_flowrate_df_molh(self, value):
        self.molar_flowrate_df_molh_cache = value

    def __init__(self):
        from catrxneng.quantities import MolarFlowRate

        self.aggregate_flow_rates: dict[str, MolarFlowRate] = {}

    def check_components(self):
        if self.p0.size != len(self.kinetic_model.COMPONENTS):
            raise ValueError(
                "Number of components for reactor and rate model do not match."
            )

    def plot(self, x, y, names, modes, xlabel, ylabel, title=None):
        fig = go.Figure()
        for i in range(len(y)):
            trace = go.Scatter(x=x[i], y=[i], mode=modes[i], name=names[i])
            fig.add_trace(trace)
        fig.update_layout(
            title=dict(text=f"<b>{title}</b>", x=0.5),
            xaxis_title=f"<b>{xlabel}</b>",
            yaxis_title=f"<b>{ylabel}</b>",
            width=800,
        )
        fig.show()



from .species import Species
from .co import CO
from .h2o import H2O
from .co2 import CO2
from .h2 import H2
from .c2h4 import C2H4
from .ch4 import CH4
from .n2 import N2
from .nh3 import NH3
from .ar import Ar
from .inert import Inert
from .o2 import O2
from .graphite import Graphite
from .ch3oh import CH3OH
from .gas_mixture import GasMixture
from .c2h6 import C2H6
from .c3h6 import C3H6
from .c3h8 import C3H8
from .c4h10 import C4H10
from .c4h8 import C4H8
from .c5h10 import C5H10
from .c5h12 import C5H12
from .dme import DME

# Build a case-insensitive mapping of class names to class objects for only those imported here
import inspect

CLASS_MAP = {}
for name, obj in list(locals().items()):
    if inspect.isclass(obj):
        CLASS_MAP[name.lower()] = obj


def get_dict_from_list(comp_list):
    return {comp_id: CLASS_MAP[comp_id] for comp_id in comp_list}



import numpy as np
from ..reverse_kinetic_model import ReverseKineticModel
from .ghosh2021_no_h2_in_denom import Ghosh2021NoH2InDenom


class Ghosh2021WithoutSabatierRev(ReverseKineticModel):
    FORWARD_MODEL = Ghosh2021NoH2InDenom
    COMPONENTS = FORWARD_MODEL.COMPONENTS
    REACTIONS = FORWARD_MODEL.REACTIONS
    LIMITING_REACTANT = "ch3oh"
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    STOICH_COEFF = -1.0 * FORWARD_MODEL.STOICH_COEFF

    @property
    def Keq(self):
        return 1.0 / self.forward_model.Keq



import numpy as np
from numpy.typing import NDArray

from ..kinetic_model import KineticModel
import catrxneng.utils.equations as eqn
from catrxneng import reactions
import catrxneng.species as species
import catrxneng.quantities as quant


class CoToMeohWithDehydration(KineticModel):
    LIMITING_REACTANT = "co"
    T_REF = quant.Temperature(C=400)
    REACTANTS = {
        "co": species.CO2,
        "h2": species.H2,
    }
    PRODUCTS = {
        "ch3oh": species.CH3OH,
        "h2o": species.H2O,
        "dme": species.DME,
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    REACTIONS = {
        "co_to_meoh": reactions.CoToMeoh,
        "meoh_dehydration": reactions.MeohDehydration,
    }
    # ORDER = np.array([0.78, 1.5])
    STOICH_COEFF = np.array(
        [
            [-1, -2, 1, 0, 0, 0],
            [0, 0, -2, 1, 1, 0],
        ]
    )
    # KREF_MOLSKGCATBAR = np.array([0.000373, 1.8e-3])
    # EA_KJMOL = np.array([100.1, 54.5])

    def __init__(
        self,
        T: quant.Temperature | None = None,
        kref: NDArray | None = None,
        Ea: quant.Energy | None = None,
    ):
        self.kref = kref
        # if self.kref is None:
        #     self.kref = self.KREF_MOLSKGCATBAR
        self.Ea = Ea
        # if self.Ea is None:
        #     self.Ea = self.EA_KJMOL
        super().__init__(T)

    def compute_temp_dependent_constants(self):
        self.Keq_co_to_meoh = reactions.CoToMeoh.Keq(self.T)
        self.Keq_meoh_dehydration = reactions.MeohDehydration.Keq(self.T)
        self.Keq = np.array([self.Keq_co_to_meoh, self.Keq_meoh_dehydration])
        # self.k = np.array(
        #     [
        #         quant.RateConstant(
        #             molskgcatbar=kref,
        #             Ea=quant.Energy(kJmol=Ea),
        #             Tref=self.T_REF,
        #             order=order,
        #         )(self.T).molhgcatbar
        #         for kref, Ea, order in zip(self.kref, self.Ea, self.ORDER)
        #     ]
        # )

    # def calculate_rates(self, p: quant.Pressure) -> quant.ReactionRate:
    #     return quant.ReactionRate(
    #         molhgcat=self.rate_equations(p_array=p.bar), keys=self.comp_list()
    #     )

    # def rate_equations(self, p_array: np.ndarray) -> NDArray:
    #     """
    #     Calculate reaction rates from partial pressures.

    #     Pressure in bar
    #     Rates in mol/h/gcat

    #     Parameters
    #     ----------
    #     p_array : array-like
    #         Partial pressures. Can be:
    #         - 1D array of shape (7,) for a single point
    #         - 2D array of shape (7, n) for n points

    #     Returns
    #     -------
    #     rates : ndarray
    #         Reaction rates with the same shape as input.
    #         - 1D array of shape (7,) if input is 1D
    #         - 2D array of shape (7, n) if input is 2D
    #     """
    #     p_co2 = p_array[0]  # co2
    #     p_h2 = p_array[1]  # h2
    #     p_ch3oh = p_array[2]  # ch3oh
    #     p_h2o = p_array[3]  # h2o
    #     p_co = p_array[4]  # co

    #     base = 1 + self.Kads_co2 * p_co2 + np.sqrt(self.Kads_h2 * p_h2)
    #     inhib = base * base

    #     # Reaction 1: CO2 + 3H2 -> CH3OH + H2O (CO2-to-MeOH)
    #     p_h2_3 = p_h2 * p_h2 * p_h2
    #     beta_co2_to_meoh = 1 / self.Keq_co2_to_meoh * p_ch3oh * p_h2o / p_co2 / p_h2_3
    #     r1 = self.k[0] * p_h2**0.91 * p_co2**-0.13 * (1 - beta_co2_to_meoh)

    #     # Reaction 2: CO2 + H2 -> CO + H2O (RWGS)
    #     fwd = p_co2 * p_h2
    #     rev = p_co * p_h2o / self.Keq_rwgs
    #     numerator = self.k[1] * (fwd - rev) / np.sqrt(p_h2)
    #     r2 = numerator / inhib

    #     return np.array(
    #         [
    #             -r1 - r2,  # co2
    #             -3 * r1 - r2,  # h2
    #             r1,  # ch3oh
    #             r1 + r2,  # h2o
    #             r2,  # co
    #             0.0 * r1,  # inert
    #         ]
    #     )



import scipy.integrate as integrate

from .reaction import Reaction
from .. import species
from .. import quantities as quant


class CoToMeoh(Reaction):
    EQUATION = "CO + 2H₂ ⇌ CH₃OH"
    COMPONENTS = {
        "co": species.CO,
        "h2": species.H2,
        "ch3oh": species.CH3OH,
        "inert": species.Ar,
    }
    STOICH_COEFF = quant.Dimensionless(
        si=[-1.0, -2.0, 1.0, 0.0], keys=list(COMPONENTS.keys())
    )
    DEFAULT_LIMITING_REACTANT = "co"
    DH_RXN_298 = (
        species.CH3OH.HF_298_LIQ - species.CO.HF_298_GAS - 2 * species.H2.HF_298_GAS
    )
    DS_RXN_298 = (
        species.CH3OH.S_298_LIQ - species.CO.S_298_GAS - 2 * species.H2.S_298_GAS
    )

    @staticmethod
    def dCp_1(T_K):
        T = quant.Temperature(K=T_K)
        dCp = species.CH3OH.Cp_liq(T) - species.CO.Cp_gas(T) - 2 * species.H2.Cp_gas(T)
        return dCp.JmolK

    @staticmethod
    def dCp_2(T_K):
        T = quant.Temperature(K=T_K)
        dCp = species.CH3OH.Cp_gas(T) - species.CO2.Cp_gas(T) - 2 * species.H2.Cp_gas(T)
        return dCp.JmolK

    @classmethod
    def dH_rxn(cls, T):
        Tb_ch3oh = species.CH3OH.BOILING_TEMP.K
        if T.K < Tb_ch3oh:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, T.K)[0]
            return quant.Energy(Jmol=dHr)
        else:
            dHr = cls.DH_RXN_298.Jmol
            dHr += integrate.quad(cls.dCp_1, 298, Tb_ch3oh)[0]
            dHr += species.CH3OH.DH_VAP.Jmol
            dHr += integrate.quad(cls.dCp_2, Tb_ch3oh, T.K)[0]
            return quant.Energy(Jmol=dHr)

    @classmethod
    def dS_rxn(cls, T):
        Tb_ch3oh = species.CH3OH.BOILING_TEMP.K

        def integrand1(T_K):
            return cls.dCp_1(T_K) / T_K

        def integrand2(T_K):
            return cls.dCp_2(T_K) / T_K

        if T.K < Tb_ch3oh:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, T.K)[0]
            return quant.Entropy(JmolK=dSr)
        else:
            dSr = cls.DS_RXN_298.JmolK
            dSr += integrate.quad(integrand1, 298, Tb_ch3oh)[0]
            dSr += species.CH3OH.DS_VAP.JmolK
            dSr += integrate.quad(integrand2, Tb_ch3oh, T.K)[0]
            return quant.Entropy(JmolK=dSr)



from .species import Species
from ..quantities import Energy, Entropy, Temperature


class CH3OH(Species):
    CLASS = "alcohol"
    C_ATOMS = 1
    H_ATOMS = 4
    O_ATOMS = 1
    MOL_WEIGHT = 32
    HF_298_LIQ = Energy(kJmol=-238.4)
    S_298_LIQ = Entropy(JmolK=127.19)
    HF_298_GAS = Energy(kJmol=-205)
    S_298_GAS = Entropy(JmolK=239.7)
    DH_VAP = HF_298_GAS - HF_298_LIQ
    DS_VAP = S_298_GAS - S_298_LIQ
    # DH_VAP = Energy(kJmol=37.6)
    # DS_VAP = Entropy(JmolK=104.6)
    BOILING_TEMP = Temperature(C=64.7)

    # def __init__(self, T=None):
    #     super().__init__(T)

    # @classmethod
    # def Hf_gas(cls, T: Temperature) -> Energy:
    #     return cls.HF_298_GAS

    # @classmethod
    # def S_gas(cls, T: Temperature) -> Entropy:
    #     return cls.S_298_GAS

    @staticmethod
    def Cp_liq(T):
        from ..quantities import HeatCapacity

        # if not T:
        #     T = self.T
        cp = 0.1955 * T.K + 22.419
        return HeatCapacity(JmolK=cp)

    @staticmethod
    def Cp_gas(T):
        from ..quantities import HeatCapacity

        # if not T:
        #     T = self.T
        cp = 0.0656 * T.K + 26.21
        return HeatCapacity(JmolK=cp)



from .quantity import Quantity
from catrxneng.utils import *


class InvWHSV(Quantity):

    def __init__(self, avg_mol_weight=None, **kwargs):
        super().__init__(**kwargs)
        self.avg_mol_weight = avg_mol_weight

    @property
    def molskgcat(self):
        return self.si

    @molskgcat.setter
    def molskgcat(self, value):
        self.si = self.to_float(value)

    @property
    def molhgcat(self):
        return self.si / 3600 * 1000

    @molhgcat.setter
    def molhgcat(self, value):
        self.si = self.to_float(value) * 3600 / 1000

    @property
    def smLhgcat(self):
        return self.si / 3600 / 22.4

    @smLhgcat.setter
    def smLhgcat(self, value):
        self.si = self.to_float(value) * 3600 * 22.4

    @property
    def inv_h(self):
        try:
            return self.si / 1000 * self.avg_mol_weight * 3600
        except TypeError:
            raise AttributeError("WHSV has no avg_mol_weight assigned.")

    @inv_h.setter
    def inv_h(self, value):
        try:
            self.si = value * 1000 / self.avg_mol_weight / 3600
        except TypeError:
            raise AttributeError("WHSV has no avg_mol_weight assigned.")



from .species import Species
import catrxneng.quantities as quant


class DME(Species):
    CLASS = "ether"
    C_ATOMS = 2
    H_ATOMS = 6
    O_ATOMS = 1
    MOL_WEIGHT = 46.069
    HF_298_GAS = quant.Energy(kJmol=-184.1)
    S_298_GAS = quant.Entropy(JmolK=266.5)
    # S_298_LIQ = quant.Entropy(JmolK=146.57)
    # DS_VAP = quant.Entropy(JmolK=86.61)
    # S_298_GAS = S_298_LIQ + DS_VAP

    # @classmethod
    # def Hf_gas(cls, T: quant.Temperature) -> quant.Energy:
    #     return cls.HF_298_GAS

    # @classmethod
    # def S_gas(cls, T: quant.Temperature) -> quant.Entropy:
    #     return cls.S_298_GAS

    @classmethod
    def Cp_gas(cls, T: quant.Temperature) -> quant.HeatCapacity:
        if 298 <= T.K <= 1000:
            cp = -5e-5 * T.K * T.K + 0.173 * T.K + 18.717
            return quant.HeatCapacity(JmolK=cp)
        raise ValueError("Temperature is out of range.")



import numpy as np

import catrxneng.species as species
from ..composite_kinetic_model import CompositeKineticModel
from ..co2_to_meoh import Ghosh2021
from ..mto import Ghosh2022
import catrxneng.quantities as quant


class Ghosh2022Composite(CompositeKineticModel):
    LIMITING_REACTANT = "co2"
    REACTANTS = {
        "co2": species.CO2,
        "h2": species.H2,
    }
    PRODUCTS = {
        "ch3oh": species.CH3OH,
        "h2o": species.H2O,
        "co": species.CO,
        "ch4": species.CH4,
        "c2h4": species.C2H4,
        "c3h6": species.C3H6,
        "c4h8": species.C4H8,
        "c2h6": species.C2H6,
        "c3h8": species.C3H8,
        "c4h10": species.C4H10,
        "c58": species.Species(C_ATOMS=6.5),
        "c9+": species.Species(C_ATOMS=10.5),
    }
    COMPONENTS = {**REACTANTS, **PRODUCTS, "inert": species.Inert}
    C_ATOMS = np.array([comp.C_ATOMS for comp in COMPONENTS.values()])
    KINETIC_MODEL_CLASSES = [Ghosh2021, Ghosh2022]
    REACTIONS = {
        id: REACTION_CLASS
        for KM in KINETIC_MODEL_CLASSES
        for id, REACTION_CLASS in KM.REACTIONS.items()
    }

    def calculate_rates(self, p: quant.Pressure):
        return quant.ReactionRate(
            molhgcat=self.rate_equations(p_array=p.bar), keys=self.comp_list()
        )

    def rate_equations(self, p_array):
        map = self.map_child_components_to_parent_components
        r1 = self.kinetic_models[0].rate_equations(p_array[map[0]])
        r1 = r1 * self.catalyst_frac[0]
        r2 = self.kinetic_models[1].rate_equations(p_array[map[1]])
        r2 = r2 * self.catalyst_frac[1]
        return np.array(
            [
                r1[0],  # co2
                r1[1] + r2[1],  # h2
                r1[2] + r2[0],  # ch3oh
                r1[3] + r2[2],  # h2o
                r1[4],  # co
                r1[5],  # ch4
                r2[3],  # c2h4
                r2[4],  # c3h6
                r2[5],  # c4h8
                r2[6],  # c2h6
                r2[7],  # c3h8
                r2[8],  # c4h10
                r2[9],  # c58
                r2[10],  # c9+
                0.0 * r1[0],  # inert
            ]
        )



from .quantity import Quantity
from catrxneng.utils import *
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..species import GasMixture


class WHSV(Quantity):
    def __init__(self, gas_mixture: "GasMixture" = None, **kwargs):
        self.gas_mixture = gas_mixture
        super().__init__(**kwargs)

    @property
    def molskgcat(self):
        return self.si

    @molskgcat.setter
    def molskgcat(self, value):
        self.si = self.to_float(value)

    @property
    def molhgcat(self):
        return self.si * 3.6

    @molhgcat.setter
    def molhgcat(self, value):
        self.si = self.to_float(value) / 3.6

    @property
    def smLhgcat(self):
        return self.si * 3600 * 22.4

    @smLhgcat.setter
    def smLhgcat(self, value):
        self.si = self.to_float(value) / 3600 / 22.4

    @property
    def inv_h(self):
        try:
            return self.si / 1000 * self.gas_mixture.avg_mol_weight * 3600
        except AttributeError:
            raise AttributeError("WHSV has no gas mixture assigned.")

    @inv_h.setter
    def inv_h(self, value):
        try:
            self.si = value * 1000 / self.gas_mixture.avg_mol_weight / 3600
        except TypeError:
            raise AttributeError("WHSV has no gas mixture assigned.")

    @property
    def inv_s(self):
        try:
            return self.si / 1000 * self.gas_mixture.avg_mol_weight
        except TypeError:
            raise AttributeError("WHSV has no gas mixture assigned.")

    @inv_s.setter
    def inv_s(self, value):
        try:
            self.si = value * 1000 / self.gas_mixture.avg_mol_weight
        except TypeError:
            raise AttributeError("WHSV has no gax mixture assigned.")

    def __mul__(self, other):
        from .mass import Mass
        from .molar_flow_rate import MolarFlowRate

        if isinstance(other, Mass):
            si = self.si * other.si
            return MolarFlowRate(si=si)
        return super().__mul__(other)

    def __rmul__(self, other):
        from .mass import Mass
        from .molar_flow_rate import MolarFlowRate

        if isinstance(other, Mass):
            si = self.si * other.si
            return MolarFlowRate(si=si)
        return super().__rmul__(other)



from .species import Species
from ..quantities import *


class Ar(Species):
    CLASS = "noble_gas"
    AR_ATOMS = 1
    MOL_WEIGHT = 40
    MIN_TEMP = Temperature(K=500)
    MAX_TEMP = Temperature(K=2000)
    HF_298_GAS = Energy(kJmol=0)
    THERMO_PARAMS = {
        "A": 19.50583,
        "B": 19.88705,
        "C": -8.598535,
        "D": 1.369784,
        "E": 0.527601,
        "F": -4.935202,
        "G": 212.3900,
        "H": 0.0,
    }
